﻿// Dependencies
//#region Dependencies

// External (Common) dependencies
var moment = require('moment');
var fs = require('fs');
var cp = require('child_process');

// External (Device) dependencies
var serialport = require("serialport");
var ieee754 = require('ieee754');
var Parser = require('expr-eval').Parser;// Library to manage parsing and evaluation of formulas
var mathparser = new Parser();

// Local dependencies
var Logger;// Used for logging
var SocketIo;// Used for messaging
var NetSnmp;// Used for SNMP messaging - base library
var SmtpManager;// Used for SMTP messaging
var SnmpHostSessionsArray = [];// Instance array for list of SNMP host sessions

//#endregion Dependencies

// Constants
//#region Constants

// Constants (Common)
var DEVICE_MODEL = "./DirectionalPower+VSWR.json";//  Device model defines the device incl. general properties, parameters and I/O
var PEAK_DEVICE_SETTINGS = "./data/device"
var DELAY_DEVICEINFOCOMMANDS = 2000;
var DELAY_DEVICEINFO = 30000;
var DELAY_PROCESSING = 40000;

//#region Event Log
var EVENTS_LENGTH = 10;
var ThresholdValues = ["NORMAL", "BELOW", "ABOVE"];
//#endregion Event Log

// Power Control
//#region Power Control
//  Power control offset will be determined by selected port #
var POWER_CONTROL_LOCATION = '/sys/class/gpio/gpio';
var POWER_CONTROL_BASEGPIO = 496;
//#endregion Power Control

// Constants (Device)

//#endregion Constants

// Device Class: Variables, Instantiation, Methods and Constructor
//#region Device Class: Variables, Instantiation, Methods and Constructor
function Device(port, logger, io, smtpManager, netSnmp) {

    // Class Variables
    //#region Class Variables

    // Class Variables (Common)
    //#region Class Variables (Common)
    var EventLog = [];
    var EnabledFlag = false;
    var SendEventsFlag = false;
    var DeviceModel;
    var ThisDeviceObject = this;
    //#endregion Class Variables (Common)

    //#endregion Class Variables

    // Class Instantiation
    //#region Class Instantiation

    // Local Dependencies
    //#region Local Dependencies
    Logger = logger;
    SocketIo = io;
    SmtpManager = smtpManager;
    NetSnmp = netSnmp;
    //#endregion Local Dependencies

    try {
        // Load a private model for this instance of the device
        //  Remove model from cache to avoid overwrites
        var cache = require.resolve(DEVICE_MODEL);
        if (cache) delete require.cache[cache];
        //  Load model
        var DeviceModel = require(DEVICE_MODEL);
        //  Assign statics to model
        DeviceModel.port = port;
        DeviceModel.uuid = parseInt(moment().format("x"));
        Logger.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".Constructor: Device uuid=" + DeviceModel.uuid + " ------------------------------------------------------------");
        Logger.trace(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".Constructor: DeviceModel=" + JSON.stringify(DeviceModel));
    }
    catch (error) {
        Logger.error("Failed constructing device on port=" + port + " error=" + error);
    }
    //#endregion Class Instantiation

    // Class Methods
    //#region Class Methods

    //#region Class Methods (Common)

    //#region Class Methods (Common - Basic)
    this.setPort = function (port) {
        DeviceModel.port = port;
    }

    this.getPort = function () {
        return DeviceModel.port;
    }

    this.getDeviceId = function () {
        Logger.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".getDeviceId: Device id=" + DeviceModel.id);
        return DeviceModel.id;
    }

    this.getDeviceModel = function () {
        Logger.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".getDeviceModel: Device uuid=" + DeviceModel.uuid);
        return DeviceModel;
    }

    this.enable = function (enabled) {
        Logger.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".Enable: Enabled=" + (enabled ? "YES" : "NO"));
        if (enabled == EnabledFlag)
            return;// Don't do anything if we're already in the proper state
        else
            EnabledFlag = enabled;// Remember new setting

        if (enabled) {
            //#region Enable Power
            var power;
            if (DeviceModel.io.powerControl.enabled) power = 1; else power = 0;
            fs.truncate(POWER_CONTROL_LOCATION + (POWER_CONTROL_BASEGPIO + parseInt(DeviceModel.port)) + "/value", 0, function () {
                fs.writeFile(POWER_CONTROL_LOCATION + (POWER_CONTROL_BASEGPIO + parseInt(DeviceModel.port)) + "/value", power, function (err) {
                    if (err) {
                        return Logger.error(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".Enable: Power Set Error (Write)=" + JSON.stringify(err));
                    }
                    Logger.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".Enable: Port=" + DeviceModel.port + " - '" + POWER_CONTROL_LOCATION + (POWER_CONTROL_BASEGPIO + parseInt(DeviceModel.port)) + "/value" + "' Power=" + power);
                });
            });
            //#endregion Enable Power

            //#region Enable Serial Port
            if (!Bird4043SerialPort.isOpen()) {
                Bird4043SerialPort.open(function (error) {
                    if (error) {
                        Logger.warn(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".enable[true]: Serial Port open error=" + error);
                    } else {
                        Logger.warn(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".enable[true]: Serial Port opened");

                        // Get device information
                        setTimeout(function () { ThisDeviceObject.getDeviceInformationFromDevice() }, DELAY_DEVICEINFO);// Give this device 30 seconds to bootup

                        // Start processing 5seconds later after we get device information
                        setTimeout(function () { ThisDeviceObject.startProcessing() }, DELAY_PROCESSING);// Wait another 10 seconds until we've got the Device Info
                    }
                });
            } else {
                // Serial port is already open
                Logger.warn(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".enable[true]: Serial Port is already open=" + error);

                // Get device information
                setTimeout(function () { ThisDeviceObject.getDeviceInformationFromDevice() }, DELAY_DEVICEINFO);// Give this device 30 seconds to bootup

                // Start processing 5seconds later after we get device information
                setTimeout(function () { ThisDeviceObject.startProcessing() }, DELAY_PROCESSING);// Wait another 10 seconds until we've got the Device Info
            }
            //#endregion Enable Serial Port
        }

        else {
            this.stopProcessing();// Stop serial port commands

            //#region Disable Power
            var power;
            if (DeviceModel.io.powerControl.enabled) power = 0; else power = 1;
            fs.truncate(POWER_CONTROL_LOCATION + (POWER_CONTROL_BASEGPIO + parseInt(DeviceModel.port)) + "/value", 0, function () {
                fs.writeFile(POWER_CONTROL_LOCATION + (POWER_CONTROL_BASEGPIO + parseInt(DeviceModel.port)) + "/value", power, function (err) {
                    if (err) {
                        return Logger.error(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".Constructor: Power Set Error (Write)=" + JSON.stringify(err));
                    }
                    Logger.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".Constructor: Port=" + DeviceModel.port + " - '" + POWER_CONTROL_LOCATION + (POWER_CONTROL_BASEGPIO + parseInt(DeviceModel.port)) + "/value" + "' Power=" + power);
                });
            });
            //#endregion Disable Power

            //#region Disable Serial Port
            if (Bird4043SerialPort.isOpen()) {
                Bird4043SerialPort.close(function (error) {
                    if (error) {
                        Logger.error(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".enable[false]: Serial Port close error=" + error);
                    } else {
                        Logger.warn(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".enable[false]: Serial Port closed!");
                    }
                });
            }
            //#endregion Disable Serial Port
        }
    }

    this.sendEvents = function (sendevents) {
        SendEventsFlag = sendevents;
    }
    //#endregion Class Methods (Common - Basic)

    //#region Class Methods (Common - UpdateDevice)
    this.updateDevice = function (port, model, callback) {
        Logger.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".updateDevice: Device uuid=" + DeviceModel.uuid);
        DeviceModel = model;
        callback(DeviceModel);
    }
    //#endregion Class Methods (Common - UpdateDevice)

    //#endregion Class Methods (Common)

    //#region Class Methods (Device Processing)

    //#region Class Methods (Device Processing - Start/Stop)
    //#region startProcessing
    var GetParametersCycleControl;
    this.startProcessing = function () {
        //  Start Serial IO - Get Parameter loop
        //#region Start Serial IO - Get Parameter loop
        GetParametersCycleControl = setInterval(function () { ThisDeviceObject.Bird4043GetParameters() }, DeviceModel.io.serial.cycletime);
        Logger.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".startProcessing: IO processing cycle=" + DeviceModel.io.serial.cycletime + "ms");
        //#endregion Start Serial IO - Get Parameter loop
    }
    //#endregion startProcessing

    //#region stopProcessing
    this.stopProcessing = function () {

        //  Stop serial IO processing loops
        //#region  Stop serial IO processing loops
        try {
            clearInterval(DeviceInfoControl);
            Logger.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".stopProcessing: Stopping get Device Information cycle.");
            clearInterval(GetParametersCycleControl);
            Logger.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".stopProcessing: Stopping IO processing cycle.");
        } catch (error) {
            Logger.error(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".stopProcessing: Error stopping DeviceInfo and IO processing cycle. Error=" + error);
        } finally {
            Logger.warn(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".stopProcessing: Stopped DeviceInfo and IO processing cycle");
        }
        //#endregion  Stop serial IO processing loops
    }
    //#endregion stopProcessing

    //#region Class Methods (Device Processing - Device Information)
    //#region getDeviceInformation
    var DeviceInfoControl;
    this.getDeviceInformationFromDevice = function () {
        // Serial Cycle
        DeviceInfoControl = setInterval(function () { ThisDeviceObject.Bird4043GetDeviceInfo() }, DELAY_DEVICEINFOCOMMANDS);
        Logger.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".getDeviceInformationFromDevice: Get Device Info started");
    }

    this.getDeviceInformation = function () {
        return DeviceModel.device;
    }
    //#endregion getDeviceInformation

    //#endregion Class Methods (Device Processing - Device Information)

    //#endregion Class Methods (Device Processing - Start/Stop)

    //#region Class Methods (Device Processing - Process I/O)

    //#region Serial I/O
    //  Reads Forward Power, Reflected Power, VSWR from Bird 4043
    //  Updates UI based on I/O changes
    var Bird4043SendBuffer = new Uint8Array(5);
    var Bird4043DeviceInfoCommands = [2, 4, 16, 43]
    var Bird4043DeviceInfoCommandsString = ["Get Serial Number", "Get Device Name", "Get Model Number", "Get Firmware Version"];
    var Bird4043DeviceInfoReturnLength = [9, 25, 23, 7]
    var Bird4043DeviceInfoCommandsIndex = 0;
    var Bird4043ParameterCommands = [10, 11, 12, 29]
    var Bird4043ParameterReturnLength = [9, 9, 9, 9]
    var Bird4043ParameterCommandsIndex = 0;
    var ExpectedResponseLength = 0;
    var CurrentResponseBufferPtr = 0;
    var ResponseBuffer = new Buffer(256);
    var Bird4043SerialPort = new serialport("/dev/ttyUSB" + DeviceModel.port, { baudRate: 57600, dataBits: 8, stopBits: 1, parity: 'none', parser: serialport.parsers.raw, autoOpen: false});

    // Bird 4043
    // Create and open serial port
    Bird4043SerialPort.on('open', function () {
        Logger.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".Bird4043SerialPort: Opened");
    });

    this.Bird4043GetDeviceInfo = function () {
        // Check to see if we've been disabled
        if (!EnabledFlag) {
            try {
                clearInterval(DeviceInfoControl);
                Logger.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".Bird4043GetParameters: Stopping GetDeviceInfo (DeviceInfoControl)");
            }
            catch (error) {
                Logger.error(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".Bird4043GetParameters: Error stopping GetDeviceInfo (DeviceInfoControl) error=" + error);
            }
            finally {
                Logger.warn(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".Bird4043GetParameters: Stopped GetDeviceInfo (DeviceInfoControl)");
                return;
            }
        }

        // Send next command in the cycle
        //#region Send next command in the cycle
        Bird4043SendBuffer[0] = DeviceModel.device.DeviceAddress;  // Device Addrsess - for this release we'll use default value in JSON file (probably 255)
        Bird4043SendBuffer[1] = Bird4043DeviceInfoCommands[Bird4043DeviceInfoCommandsIndex];
        Bird4043SendBuffer[2] = 0;
        crc = this.CalculateCRC(Bird4043SendBuffer, 3);
        Bird4043SendBuffer[3] = (crc >> 8) & 0xFF;
        Bird4043SendBuffer[4] = crc & 0xFF;

        // Set how much data we expect to receive
        ExpectedResponseLength = Bird4043DeviceInfoReturnLength[Bird4043DeviceInfoCommandsIndex];
        Logger.info("this.Bird4043GetDeviceInfo: Command=" + Bird4043DeviceInfoCommands[Bird4043DeviceInfoCommandsIndex] + " (" + Bird4043DeviceInfoCommandsString[Bird4043DeviceInfoCommandsIndex] + ")");

        //for (var i = 0; i < 5; i++) {
        //    logger.info("this.Bird4043GetDeviceInfo: Sending Buffer[" + i + "]=" + Bird4043SendBuffer[i]);
        //}

        // Stop processing once we're done the command in the list
        if (++Bird4043DeviceInfoCommandsIndex == Bird4043DeviceInfoCommands.length) clearInterval(DeviceInfoControl);

        Bird4043SerialPort.write(Bird4043SendBuffer, function (error) {
            if (error) {
                logger.error("this.Bird4043GetDeviceInfo: Error writing to serial port");
                return;
            }
        });
        //#endregion Send next command in the cycle
    }

    this.Bird4043GetParameters = function () {
        // Check to see if we've been disabled
        if (!EnabledFlag) {
            try {
                clearInterval(GetParametersCycleControl);
                Logger.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".Bird4043GetParameters: Stopping GetParameters (GetParametersCycleControl)");
            }
            catch (error) {
                Logger.error(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".Bird4043GetParameters: Error stopping GetParameters (GetParametersCycleControl) error=" + error);
            }
            finally {
                Logger.warn(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".Bird4043GetParameters: Stopped GetParameters (GetParametersCycleControl)");
                return;
            }
        }

        // Send next command in the cycle
        //#region Send next command in the cycle
        Bird4043SendBuffer[0] = DeviceModel.device.DeviceAddress;  // Device Addrsess - for this release we'll use default value in JSON file (probably 255)
        Bird4043SendBuffer[1] = Bird4043ParameterCommands[Bird4043ParameterCommandsIndex]; // Forward Power, Reflected Power, VSWR, Temperature ... then do it again
        Bird4043SendBuffer[2] = 0;
        crc = this.CalculateCRC(Bird4043SendBuffer, 3);
        Bird4043SendBuffer[3] = (crc >> 8) & 0xFF;
        Bird4043SendBuffer[4] = crc & 0xFF;

        //for (var i = 0; i < 5; i++) {
        //    logger.info("this.Bird4043GetParameters: Sending Buffer[" + i + "]=" + Bird4043SendBuffer[i]);
        //}

        // Reset the index
        if (++Bird4043ParameterCommandsIndex == Bird4043ParameterCommands.length) Bird4043ParameterCommandsIndex = 0;

        Bird4043SerialPort.write(Bird4043SendBuffer, function (error) {
            if (error) {
                logger.error("this.Bird4043GetParameters: Error writing to serial port");
                return;
            }
        });
        //#endregion Send next command in the cycle
    }

    Bird4043SerialPort.on('data', function (data) {        
        var timeChanged = moment();
        var _buffer;// Temporary buffer
        var _ieee754Value;// IEEE754 calculated value
        var _4BytesValue;// 4-byte little endian value
        var BufferString;// Human view of buffer
        var _expression;// Formula expression value 

        // If we have full data then process, else append and wait for more
        //Logger.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".Bird4043SerialPort.data: Received=" + data.toString('hex') + " (Rx=" + data.length + ", Expected=" + ExpectedResponseLength + "), CurrentPtr=" + CurrentResponseBufferPtr);
        data.copy(ResponseBuffer, CurrentResponseBufferPtr, 0, data.length);
        CurrentResponseBufferPtr += data.length;
        if (CurrentResponseBufferPtr < ExpectedResponseLength) {
            // Copy received buffer into response, and then wait for more data
            //Logger.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".Bird4043SerialPort.data: Partial=" + data.toString('hex') + " (Rx=" + data.length + ", Expected=" + ExpectedResponseLength + ")");

            // Check for adjusted Expected response length
            if (CurrentResponseBufferPtr >= 2) {
                if (ResponseBuffer[2] != (ExpectedResponseLength-5)) {
                    ExpectedResponseLength = ResponseBuffer[2] + 5;
                    //Logger.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".Bird4043SerialPort.data: Adjusted ExpectedLength=" + ExpectedResponseLength);
                }                
            }
            return;
        } else {
            // We've read the full length so go back to beginning of buffer
            //Logger.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".Bird4043SerialPort.data: Full=" + ResponseBuffer.toString('hex') + " (Rx=" + CurrentResponseBufferPtr + ", Expected=" + ExpectedResponseLength + ")");
            CurrentResponseBufferPtr = 0;
        }

        // Parse the response
        switch (ResponseBuffer[1]) {
            case 2:// SerialNumber
                _4BytesValue = (ResponseBuffer[6] << 24) + (ResponseBuffer[5] << 16) + (ResponseBuffer[4] << 8) + (ResponseBuffer[3])
                DeviceModel.device.SerialNumber = _4BytesValue;
                BufferString = (ResponseBuffer.toString('hex')).substr(0, (ResponseBuffer[2] + 5)*2);
                Logger.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".Bird4043SerialPort.ResponseBuffer: SerialNumber=" + DeviceModel.device.SerialNumber + " (Buffer=" + BufferString + ")");
                ThisDeviceObject.sendSocket(DeviceModel.port, DeviceModel.uuid, "DeviceInformation", { "DeviceInformation": DeviceModel.device });
                break;
            case 4:// Name
                var _NameAsciiBuffer = new Buffer(ExpectedResponseLength - 5);
                ResponseBuffer.copy(_NameAsciiBuffer, 0, 3, (ExpectedResponseLength - 2));
                DeviceModel.device.Name = _NameAsciiBuffer.toString('utf8');
                BufferString = (ResponseBuffer.toString('hex')).substr(0, (ResponseBuffer[2] + 5) * 2);
                Logger.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".Bird4043SerialPort.ResponseBuffer: DeviceName=" + DeviceModel.device.Name + " (Buffer=" + BufferString + ")");
                ThisDeviceObject.sendSocket(DeviceModel.port, DeviceModel.uuid, "DeviceInformation", { "DeviceInformation": DeviceModel.device });
                break;
            case 16:// Model
                var _ModelAsciiBuffer = new Buffer(ExpectedResponseLength - 5);
                ResponseBuffer.copy(_ModelAsciiBuffer, 0, 3, (ExpectedResponseLength - 2));
                DeviceModel.device.Model = _ModelAsciiBuffer.toString('utf8');
                BufferString = (ResponseBuffer.toString('hex')).substr(0, (ResponseBuffer[2] + 5) * 2);
                Logger.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".Bird4043SerialPort.ResponseBuffer: Model=" + DeviceModel.device.Model + " (Buffer=" + BufferString + ")");
                ThisDeviceObject.sendSocket(DeviceModel.port, DeviceModel.uuid, "DeviceInformation", { "DeviceInformation": DeviceModel.device });
                break;
            case 43:// Firmware
                var _FirmwareAsciiBuffer = new Buffer(ExpectedResponseLength - 5);
                ResponseBuffer.copy(_FirmwareAsciiBuffer, 0, 3, (ExpectedResponseLength - 2));
                DeviceModel.device.Firmware = _FirmwareAsciiBuffer.toString('utf8');
                BufferString = (ResponseBuffer.toString('hex')).substr(0, (ResponseBuffer[2] + 5) * 2);
                Logger.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".Bird4043SerialPort.ResponseBuffer: Firmware=" + DeviceModel.device.Firmware + " (Buffer=" + BufferString + ")");
                ThisDeviceObject.sendSocket(DeviceModel.port, DeviceModel.uuid, "DeviceInformation", { "DeviceInformation": DeviceModel.device });
                ThisDeviceObject.saveDeviceModel(DeviceModel);
                break;
            case 10:// Forward Power
                _buffer = new Buffer([ResponseBuffer[3], ResponseBuffer[4], ResponseBuffer[5], ResponseBuffer[6]]);
                _ieee754Value = ieee754.read(_buffer, 0, true, 23, 4);
                if (_ieee754Value != DeviceModel.parameters.ForwardPower.RawValue) {
                    DeviceModel.parameters.ForwardPower.RawValue = _ieee754Value;
                    DeviceModel.parameters.ForwardPower.TimeChanged = timeChanged;
                    for (i = 0; i < DeviceModel.parameters.ForwardPower.Formulas.length; i++) {
                        _expression = mathparser.parse(DeviceModel.parameters.ForwardPower.Formulas[i].Formula);
                        // If Watts, add a special conversion for mWatts
                        if (DeviceModel.parameters.ForwardPower.Formulas[i].Name === "Watts") {
                            if (Number(_expression.evaluate({ x: DeviceModel.parameters.ForwardPower.RawValue })) < 1) {
                                DeviceModel.parameters.ForwardPower.Formulas[i].Units = "mW"
                                DeviceModel.parameters.ForwardPower.Formulas[i].Decimals = 1
                                DeviceModel.parameters.ForwardPower.Formulas[i].FormulaValue = Number(Number(_expression.evaluate({ x: DeviceModel.parameters.ForwardPower.RawValue })) * 1000).toFixed(DeviceModel.parameters.ForwardPower.Formulas[i].Decimals);
                            } else {
                                DeviceModel.parameters.ForwardPower.Formulas[i].Units = "W"
                                DeviceModel.parameters.ForwardPower.Formulas[i].Decimals = 2
                                DeviceModel.parameters.ForwardPower.Formulas[i].FormulaValue = Number(_expression.evaluate({ x: DeviceModel.parameters.ForwardPower.RawValue })).toFixed(DeviceModel.parameters.ForwardPower.Formulas[i].Decimals);
                            }
                        }
                        // All other formulas
                        else {
                            DeviceModel.parameters.ForwardPower.Formulas[i].FormulaValue = Number(_expression.evaluate({ x: DeviceModel.parameters.ForwardPower.RawValue })).toFixed(DeviceModel.parameters.ForwardPower.Formulas[i].Decimals);
                        } 
                        Logger.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".Bird4043SerialPort.ResponseBuffer: Forward Power=" + DeviceModel.parameters.ForwardPower.Formulas[i].FormulaValue + " " + DeviceModel.parameters.ForwardPower.Formulas[i].Units + " Raw=" + DeviceModel.parameters.ForwardPower.RawValue);
                    }
                    ThisDeviceObject.sendSocket(DeviceModel.port, DeviceModel.uuid, "ForwardPower", { "ForwardPower": DeviceModel.parameters.ForwardPower });
                }
                break;
            case 11:// Reflected Power
                _buffer = new Buffer([ResponseBuffer[3], ResponseBuffer[4], ResponseBuffer[5], ResponseBuffer[6]]);
                _ieee754Value = ieee754.read(_buffer, 0, true, 23, 4);
                if (_ieee754Value != DeviceModel.parameters.ReflectedPower.RawValue) {
                    DeviceModel.parameters.ReflectedPower.RawValue = _ieee754Value;
                    DeviceModel.parameters.ReflectedPower.TimeChanged = timeChanged;
                    for (i = 0; i < DeviceModel.parameters.ReflectedPower.Formulas.length; i++) {
                        _expression = mathparser.parse(DeviceModel.parameters.ReflectedPower.Formulas[i].Formula);
                        // If Watts, add a special conversion for mWatts
                        if (DeviceModel.parameters.ReflectedPower.Formulas[i].Name === "Watts") {
                            if (Number(_expression.evaluate({ x: DeviceModel.parameters.ReflectedPower.RawValue })) < 1) {
                                DeviceModel.parameters.ReflectedPower.Formulas[i].Units = "mW"
                                DeviceModel.parameters.ReflectedPower.Formulas[i].Decimals = 1
                                DeviceModel.parameters.ReflectedPower.Formulas[i].FormulaValue = Number(Number(_expression.evaluate({ x: DeviceModel.parameters.ReflectedPower.RawValue })) * 1000).toFixed(DeviceModel.parameters.ReflectedPower.Formulas[i].Decimals);
                            } else {
                                DeviceModel.parameters.ReflectedPower.Formulas[i].Units = "W"
                                DeviceModel.parameters.ReflectedPower.Formulas[i].Decimals = 2
                                DeviceModel.parameters.ReflectedPower.Formulas[i].FormulaValue = Number(_expression.evaluate({ x: DeviceModel.parameters.ReflectedPower.RawValue })).toFixed(DeviceModel.parameters.ReflectedPower.Formulas[i].Decimals);
                            }
                        }
                        // All other formulas
                        else {
                            DeviceModel.parameters.ReflectedPower.Formulas[i].FormulaValue = Number(_expression.evaluate({ x: DeviceModel.parameters.ReflectedPower.RawValue })).toFixed(DeviceModel.parameters.ReflectedPower.Formulas[i].Decimals);
                        }
                        Logger.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".Bird4043SerialPort.ResponseBuffer: Reflected Power=" + DeviceModel.parameters.ReflectedPower.Formulas[i].FormulaValue + " " + DeviceModel.parameters.ReflectedPower.Formulas[i].Units + " Raw=" + DeviceModel.parameters.ReflectedPower.RawValue);
                    }
                    ThisDeviceObject.sendSocket(DeviceModel.port, DeviceModel.uuid, "ReflectedPower", { "ReflectedPower": DeviceModel.parameters.ReflectedPower });
                }
                break;
            case 12:// VSWR
                _buffer = new Buffer([ResponseBuffer[3], ResponseBuffer[4], ResponseBuffer[5], ResponseBuffer[6]]);
                _ieee754Value = ieee754.read(_buffer, 0, true, 23, 4);
                if (_ieee754Value != DeviceModel.parameters.VSWR.RawValue) {
                    DeviceModel.parameters.VSWR.RawValue = Number(_ieee754Value);
                    DeviceModel.parameters.VSWR.TimeChanged = timeChanged;
                    for (i = 0; i < DeviceModel.parameters.VSWR.Formulas.length; i++) {
                        _expression = mathparser.parse(DeviceModel.parameters.ReflectedPower.Formulas[i].Formula);
                        DeviceModel.parameters.VSWR.Formulas[i].FormulaValue = Number(_expression.evaluate({ x: DeviceModel.parameters.VSWR.RawValue })).toFixed(DeviceModel.parameters.VSWR.Formulas[i].Decimals);
                        Logger.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".Bird4043SerialPort.ResponseBuffer: VSWR=" + DeviceModel.parameters.VSWR.Formulas[i].FormulaValue + " " + DeviceModel.parameters.VSWR.Formulas[i].Units + " Raw=" + DeviceModel.parameters.VSWR.RawValue);
                    }
                    ThisDeviceObject.sendSocket(DeviceModel.port, DeviceModel.uuid, "VSWR", { "VSWR": DeviceModel.parameters.VSWR });
                }
                break;
            case 29:// Temperature
                _buffer = new Buffer([ResponseBuffer[3], ResponseBuffer[4], ResponseBuffer[5], ResponseBuffer[6]]);
                _ieee754Value = ieee754.read(_buffer, 0, true, 23, 4);
                if (_ieee754Value != DeviceModel.parameters.Temperature.RawValue) {
                    DeviceModel.parameters.Temperature.RawValue = _ieee754Value;
                    DeviceModel.parameters.Temperature.TimeChanged = timeChanged;
                    for (i = 0; i < DeviceModel.parameters.Temperature.Formulas.length; i++) {
                        _expression = mathparser.parse(DeviceModel.parameters.Temperature.Formulas[i].Formula);
                        DeviceModel.parameters.Temperature.Formulas[i].FormulaValue = Number(_expression.evaluate({ x: DeviceModel.parameters.Temperature.RawValue })).toFixed(DeviceModel.parameters.Temperature.Formulas[i].Decimals);
                        Logger.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".Bird4043SerialPort.ResponseBuffer: Temperature=" + DeviceModel.parameters.Temperature.Formulas[i].FormulaValue + " " + DeviceModel.parameters.Temperature.Formulas[i].Units);
                    }
                    ThisDeviceObject.sendSocket(DeviceModel.port, DeviceModel.uuid, "Temperature", { "Temperature": DeviceModel.parameters.Temperature });
                }
                break;
        }
    });

    // Save's DeviceModel
    this.saveDeviceModel = function (DeviceModel) {
        fs.truncate(PEAK_DEVICE_SETTINGS + DeviceModel.port + ".json", 0, function () {
            fs.writeFile(PEAK_DEVICE_SETTINGS + DeviceModel.port + ".json", JSON.stringify(DeviceModel), function (error) {
                if (error) {
                    Logger.error(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".saveDeviceFile: write error '" + PEAK_DEVICE_SETTINGS + DeviceModel.port + ".json'" + error);
                }
                cp.exec("sync", function (error, stdout, stderr) {
                    Logger.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".saveDeviceFile: sync '" + PEAK_DEVICE_SETTINGS + DeviceModel.port + ".json'" + error);
                    if (error) {
                        Logger.error(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".saveDeviceFile: sync error '" + PEAK_DEVICE_SETTINGS + DeviceModel.port + ".json'" + error);
                    } else {
                        Logger.warn(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".saveDeviceFile: Device Updated! '" + PEAK_DEVICE_SETTINGS + DeviceModel.port + ".json'");
                    }
                });
            });
        });
    }

    this.CalculateCRC = function (buffer, length) {
        var crc = 0xFFFF;

        //logger.info("CalculateCRC: buffer.length=" + length);

        for (var pos = 0; pos < length; pos++) {
            crc ^= buffer[pos];
            //logger.info("CalculateCRC: pos=" + pos+ " CRC="+crc);

            for (var i = 8; i != 0; i--) {
                if ((crc & 0x0001) != 0) {
                    crc >>= 1;
                    crc ^= 0xA001;
                }
                else
                    crc >>= 1;
            }
        }
        return crc;
    }
    //#endregion Serial I/O
    //#endregion Class Methods (Device Processing - Process I/O)
    //#endregion Class Methods (Device Processing)

    //#region Class Methods (Eventing)

    //#region Class Methods (Eventing - SendEvent)
    this.sendEvent = function (port, uuid, parameter) {
        this.sendSocket(port, uuid, "Value", parameter)
        this.sendSMTP(port, uuid, parameter);
        this.sendSNMP(port, uuid, parameter);

        // EventLog
        //#region EventLog

        // Retrieve EventLog array
        if (fs.existsSync("log/port" + port + "events.json")) {
            fs.readFile("log/port" + port + "events.json", function (error, eventlog) {
                if (error) {
                    EventLog = [];// Error in EventLog retrieval
                    Logger.error(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".sendEvent: EventLog retrieve error. Error=" + error);
                } else {
                    EventLog = JSON.parse(eventlog);
                    Logger.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".sendEvent: EventLog retrieved. EventLog=" + JSON.stringify(EventLog));
                }

                // Update EventLog with latest event
                EventLog.unshift(parameter);// To the front of the array
                if (EventLog.length > EVENTS_LENGTH) {
                    EventLog.splice(-1);// Remove from the end if limit reached
                }

                // Save updated EventLog array          
                fs.truncate("log/port" + port + "events.json", 0, function () {
                    fs.writeFile("log/port" + port + "events.json", JSON.stringify(EventLog), function (error) {
                        if (error) {
                            Logger.error(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".sendEvent: EventLog save error. Error=" + error);
                        } else {
                            Logger.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".sendEvent: EventLog updated and saved. EventLog=" + JSON.stringify(EventLog));
                            cp.exec("sync", function (error, stdout, stderr) {
                                Logger.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".sendEvent: sync stdout=" + stdout);
                                if (error) {
                                    Logger.error(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".sendEvent: sync error=" + error);
                                }
                            });
                        }
                    });
                });

            });
        } else {
            // Update EventLog with latest event
            EventLog = [];// Empty event log
            EventLog.unshift(parameter);// To the front of the array

            // Save updated EventLog array
            fs.truncate("log/port" + port + "events.json", 0, function () {
                fs.writeFile("log/port" + port + "events.json", JSON.stringify(EventLog), function (error) {
                    if (error) {
                        Logger.error(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".sendEvent: EventLog save error. Error=" + error);
                    } else {
                        Logger.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".sendEvent: EventLog updated and saved. EventLog=" + JSON.stringify(EventLog));
                        cp.exec("sync", function (error, stdout, stderr) {
                            Logger.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".sendEvent: sync stdout=" + stdout);
                            if (error) {
                                Logger.error(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".sendEvent: sync error=" + error);
                            }
                        });
                    }
                });
            });
        }
        this.sendSocket(DeviceModel.port, DeviceModel.uuid, "EventLog", EventLog);// For real-time update

        //#endregion EventLog
    }
    //#endregion Class Methods (Eventing - SendEvent)

    //#region Class Methods (Eventing - Real-time Updates)
    this.sendSocket = function (port, uuid, message, parameter) {
        //Logger.warn(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".sendSocket: Port=" + port + " Message=" + message + " Parameter=" + JSON.stringify(parameter));
        SocketIo.sockets.emit(DeviceModel.id, { "Port": port, "UUID": uuid, "Message": message, "Parameter": parameter })
    }
    //#endregion Class Methods (Eventing - Real-time Updates)

    //#region Class Methods (Eventing - SMTP)
    this.sendSMTP = function (port, uuid, parameter) {

        // Send only if device's SendEventsFlag is set
        if (!SendEventsFlag) return;

        // Set up the SMTP Email message
        var message = [];
        var portNumber = (parseInt(port) + 1);
        var event = "'" + parameter.Id + "' on Port " + portNumber + " - Threshold Status=" + ThresholdValues[parameter.ThresholdState];
        message += "Timestamp=" + moment(parameter.TimeChanged).format() + "\n";
        message += "Port=" + portNumber + "\n";
        message += "Name=" + parameter.Id + "\n";
        message += "Status=" + (ThresholdValues[parameter.ThresholdState]) + "\n";
        message += "Ranged Value=" + (parameter.RangedValue) + "Volts\n";
        message += "Formula Value=" + (parameter.Formulas[parameter.FormulaIndex].FormulaValue) + "\n";
        message += "Low Threshold (Ranged)=" + (parameter.RangedLowThreshold) + "Volts\n";
        message += "Low Threshold (Formula)=" + (parameter.Formulas[parameter.FormulaIndex].FormulaLowThreshold) + "\n";
        message += "High Threshold (Ranged)=" + (parameter.RangedHighThreshold) + "Volts\n";
        message += "High Threshold (Formula)=" + (parameter.Formulas[parameter.FormulaIndex].FormulaHighThreshold) + "\n";
        message += "Hold Time=" + parseInt(parameter.HoldTime) + "\n";
        message += "Qualifier Used=" + (parameter.Qualifier ? "YES" : "NO") + "\n";

        var ioInfo = "Port: " + portNumber + " '" + parameter.Id + "'";

        Logger.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".sendEvent: Sending SMTP Event for Port " + portNumber + " '" + parameter.Id + "'");
        SmtpManager.SendSmtp(message, ioInfo, event, function (result) {
            Logger.trace(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".sendEvent: SMTP Email Event Response for Port " + portNumber + " '" + parameter.Id + "' " + result.result);
        })
    }
    //#endregion Class Methods (Eventing - SMTP)

    //#region Class Methods (Eventing - SNMP)
    this.sendSNMP = function (port, uuid, parameter) {

        // Send only if device's SendEventsFlag is set
        if (!SendEventsFlag) return;

        // Set up the SNMP trap    
        var trapOid = "1.3.6.1.4.1.11902.4." + (parseInt(port) + 1) + ".10.1.0.1";
        var varbinds = [
            {
                // Timestamp
                oid: "1.3.6.1.4.1.11902.4." + (parseInt(port) + 1) + ".1.0.1",
                type: NetSnmp.ObjectType.OctetString,
                value: new Buffer(moment().toString())
            },
            {
                // Port
                oid: "1.3.6.1.4.1.11902.4." + (parseInt(port) + 1) + ".1.0.2",
                type: NetSnmp.ObjectType.Integer,
                value: parseInt((parseInt(port) + 1))
            },
            {
                // DeviceType
                oid: "1.3.6.1.4.1.11902.4." + (parseInt(port) + 1) + ".1.0.3",
                type: NetSnmp.ObjectType.OctetString,
                value: new Buffer(DeviceModel.id)
            },
            {
                // Parameter
                oid: "1.3.6.1.4.1.11902.4." + (parseInt(port) + 1) + ".1.0.4",
                type: NetSnmp.ObjectType.OctetString,
                value: new Buffer(parameter.Id)
            },
            {
                // Value
                oid: "1.3.6.1.4.1.11902.4." + (parseInt(port) + 1) + ".1.0.5",
                type: NetSnmp.ObjectType.OctetString,
                value: new Buffer(parameter.Formulas[parameter.FormulaIndex].FormulaValue)
            },
            {
                // State
                oid: "1.3.6.1.4.1.11902.4." + (parseInt(port) + 1) + ".1.0.6",
                type: NetSnmp.ObjectType.Integer,
                value: parseInt(parameter.ThresholdState)
            },
            {
                // Units
                oid: "1.3.6.1.4.1.11902.4." + (parseInt(port) + 1) + ".1.0.7",
                type: NetSnmp.ObjectType.OctetString,
                value: new Buffer(parameter.Formulas[parameter.FormulaIndex].Units)
            }
            //,
            //{
            //    // DataString1
            //    oid: "1.3.6.1.4.1.11902.4." + (parseInt(port) + 1) + ".1.0.10",
            //    type: NetSnmp.ObjectType.OctetString,
            //    value: new Buffer(" ")
            //},
            //{
            //    // DataString2
            //    oid: "1.3.6.1.4.1.11902.4." + (parseInt(port) + 1) + ".1.0.11",
            //    type: NetSnmp.ObjectType.OctetString,
            //    value: new Buffer(" ")
            //},
            //{
            //    // DataString3
            //    oid: "1.3.6.1.4.1.11902.4." + (parseInt(port) + 1) + ".1.0.12",
            //    type: NetSnmp.ObjectType.OctetString,
            //    value: new Buffer(" ")
            //},
            //{
            //    // DataString4
            //    oid: "1.3.6.1.4.1.11902.4." + port + ".1.0.13",
            //    type: NetSnmp.ObjectType.OctetString,
            //    value: new Buffer(" ")
            //},
            //{
            //    // DataString5
            //    oid: "1.3.6.1.4.1.11902.4." + port + ".1.0.14",
            //    type: NetSnmp.ObjectType.OctetString,
            //    value: new Buffer(" ")
            //},
            //{
            //    // DataInteger1
            //    oid: "1.3.6.1.4.1.11902.4." + port + ".1.0.20",
            //    type: NetSnmp.ObjectType.Integer,
            //    value: parseInt(0)
            //},
            //{
            //    // DataInteger2
            //    oid: "1.3.6.1.4.1.11902.4." + port + ".1.0.21",
            //    type: NetSnmp.ObjectType.Integer,
            //    value: parseInt(0)
            //},
            //{
            //    // DataInteger3
            //    oid: "1.3.6.1.4.1.11902.4." + port + ".1.0.22",
            //    type: NetSnmp.ObjectType.Integer,
            //    value: parseInt(0)
            //},
            //{
            //    // DataInteger4
            //    oid: "1.3.6.1.4.1.11902.4." + port + ".1.0.23",
            //    type: NetSnmp.ObjectType.Integer,
            //    value: parseInt(0)
            //},
            //{
            //    // DataInteger5
            //    oid: "1.3.6.1.4.1.11902.4." + port + ".1.0.24",
            //    type: NetSnmp.ObjectType.Integer,
            //    value: parseInt(0)
            //}
        ];

        // Send the SNMP notification
        for (var i = 0; i < SnmpHostSessionsArray.length; i++) {
            // Skip if this host is not enabled
            if (SnmpHostSessionsArray[i].Enabled === "FALSE") continue;

            // Send trap to host ...
            var IPAddress = SnmpHostSessionsArray[i].Session.target;
            Logger.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".sendSNMP: Sending SNMP trap to IPAddress=" + IPAddress);

            SnmpHostSessionsArray[i].Session.trap(trapOid, varbinds, function (error) {
                if (error) {
                    Logger.error(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".sendSNMP: Error=" + error);
                }
                else {
                    Logger.trace(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".sendSNMP: Sending to SNMP Host @ " + IPAddress + " - Port[" + varbinds[1].value + "] '" + varbinds[2].value + "' Value=" + varbinds[4].value);
                }
            });
        }

    }
    //#endregion Class Methods (Eventing - SNMP)

    //#endregion Class Methods (Eventing)

    //#endregion Class Methods   
}

//#region Device Constructor
// Device constructor to ensure unique objects
//  Uses the following signature:
//      port        - port on which device is installed [0 indexed]
//      logger      - system logger object
//      io          - system socket io object
//      smtpManager - SMTP messaging object
//      netSnmp     - SNMP infrastructure object
module.exports = function (port, logger, io, smtpManager, netSnmp, callback) {
    var DeviceObject = new Device(port, logger, io, smtpManager, netSnmp);
    callback(DeviceObject);
}
//#endregion Device Constructor
//#region Device Class: Variables, Instantiation, Methods and Constructor

// Exported Methods
//#region Exported Methods

// Basic Info
//#region Basic Info
Device.prototype.SetPort = function (port) {
    this.setPort(port);
}

Device.prototype.GetPort = function () {
    return this.getPort()
}

Device.prototype.GetDeviceId = function () {
    return this.getDeviceId()
}

Device.prototype.GetDeviceModel = function () {
    return this.getDeviceModel()
}

Device.prototype.GetDeviceInformation = function (callback) {
    var deviceInfo = this.getDeviceInformation();
    callback(deviceInfo);
}

//#endregion Basic Info

// Enable/Disable
//#region Enable/Disable
Device.prototype.Enable = function (enabled) {
    var id = this.getDeviceId();
    Logger.trace(id + ".Enable: EnabledFlag=" + (enabled ? "YES" : "NO"));
    this.enable(enabled);
}
//#endregion Enable/Disable

// Send Events
//#region Send Events
Device.prototype.SendEvents = function (sendevents) {
    var id = this.getDeviceId();
    Logger.trace(id + ".SendEvents: SendEventsFlag=" + (sendevents ? "YES" : "NO"));
    this.sendEvents(sendevents);
}
//#endregion Send Events

// Update Device
//#region Update Device

// UpdateDevice
Device.prototype.UpdateDevice = function (port, model, callback) {
    var id = this.getDeviceId();
    this.updateDevice(port, model, function (model) {
        Logger.trace(id + ".UpdateDevice: Updated DeviceModel=" + JSON.stringify(model));
        callback(model);
    });
}

// SetSnmpHostSessionsArray
Device.prototype.SetSnmpHostSessionsArray = function (snmpHostSessions, callback) {
    var id = this.getDeviceId();
    SnmpHostSessionsArray = snmpHostSessions.slice(0);// Copy array
    Logger.trace(id + ".SetSnmpHostSessionsArray: SNMP Host Sessions established");
    callback();
}

//#endregion Update Device

//#endregion Exported Methods