// Requires
var fs = require('fs');
var moment = require('moment');
var exec = require('child_process').exec;

// Factory reset process
//  - Upon depress of the Factory Reset button, LED will change color to AMBER
//  - After FACTORYRESETIME1, any release of the Factory Reset button activates the factory reset operation, 
//    during this time, the LED will start flashing 
//  - If the Factory Reset button is kept depressed for more than FACTORYRESETTIME2, system returns back to normal

var FACTORYRESETCHECKTIME = 200;// How often we'll check for a factory reset button press
var FACTORYRESETBLINKRATE = 400;// How long the system turns off or on the LED during active mode
var FACTORYRESETENTRYTIME = 1500;// The amount of time system waits before flashing
var FACTORYRESETACTIVETIME = 6000;// The amount of time system stops flashing and returns to normal

// Factory Reset states
var FACTORYRESETOFF = 0;// Off state - this is the normal state and where we are returned if we go passed the Active state
var FACTORYRESETENTRY = 1;// This is the state before active state, where nothing happens other than LED change
var FACTORYRESETACTIVE = 2;// Factory reset is in active mode, release of button causes factory reset

var FactoryResetButtonPressedDuration = 0;// How long the button has been pressed
var FactoryResetState = FACTORYRESETOFF;// Indicates whether the button is pressed or not
var FactoryResetLED = false;// Indicates whether the LED is on (true) or off (false)

// Continual check for press of factory reset button
exports.ProcessFactoryReset = function (db, logger, io, netSnmp, ledManager, callback) {
    timeStart = moment();

    var readableStream = fs.createReadStream('/dev/input/event0');
    var data = '';

    setInterval(function () { CheckFactoryResetButtonPressedDuration(logger, io, ledManager) }, FACTORYRESETCHECKTIME);

    readableStream.on('data', function (chunk) {
        //console.log(chunk.toString('hex'));
        if (chunk[12] == 0) {
            // If we were in FACTORYRESETOFF state, 
            //  go to the FACTORYRESETENTRY state 
            if (FactoryResetState === FACTORYRESETOFF) {
                logger.info('ProcessFactoryReset: Factory Reset State = ENTRY (button pressed)');
                FactoryResetState = FACTORYRESETENTRY;

                // Turn on LED
                ledManager.WarningOperation(fs, io, logger);
                FactoryResetLED = true;
            }
        }
        else {
            if (FactoryResetState === FACTORYRESETACTIVE) {
                logger.info('ProcessFactoryReset: Factory Reset button released');
                // We are in Active state, we need to the factory reset now!

                // Set this to normal state before we leave                    
                logger.info('ProcessFactoryReset: Factory Reset has been INITIATED!!! (button released during factory reset active state)');
                FactoryResetState = FACTORYRESETOFF;
                FactoryResetButtonPressedDuration = 0;

                // Turn off LED
                ledManager.OffOperation(fs, io, logger);
                FactoryResetLED = false;

                ProcessFactoryResetOperations(logger, function (error) {
                    // We never return from this because a reboot will happen!!!
                });
            } else {
                // For any other state, just return back to normal/off state
                logger.info('ProcessFactoryReset: Factory Reset State = OFF (button released)');
                FactoryResetState = FACTORYRESETOFF;
                FactoryResetButtonPressedDuration = 0;

                // Turn off LED
                ledManager.NormalOperation(fs, io, logger);
                FactoryResetLED = false;
            }

        }
    });
};

// This function increments the states and controls LED operation
//  for advanced states
function CheckFactoryResetButtonPressedDuration(logger, io, ledManager) {

    // Increment the timer, if we are not in OFF state
    if (FactoryResetState !== FACTORYRESETOFF) {
        FactoryResetButtonPressedDuration += FACTORYRESETCHECKTIME;
    }

    //  ... then check to see if there has been a state change
    switch (FactoryResetState) {
        case FACTORYRESETOFF:
            break;

        case FACTORYRESETENTRY:
            // If we've been in this state for more than the entry time
            //  move to the active state
            if (FactoryResetButtonPressedDuration >= FACTORYRESETENTRYTIME) {
                logger.info('CheckFactoryResetButtonPressedDuration: Factory Reset State = ACTIVE (button pressed, beyond entry time = ' + FactoryResetButtonPressedDuration + 'ms)');
                FactoryResetState = FACTORYRESETACTIVE;
            }
            break;

        case FACTORYRESETACTIVE:
            // If we've been in this state for more than the entry time
            //  move to the active state
            if (FactoryResetButtonPressedDuration >= FACTORYRESETACTIVETIME) {
                logger.info('CheckFactoryResetButtonPressedDuration: Factory Reset State = OFF (button pressed, beyond active time = ' + FactoryResetButtonPressedDuration + 'ms)');
                FactoryResetState = FACTORYRESETOFF;
                FactoryResetButtonPressedDuration = 0;

                // Turn off LED
                ledManager.NormalOperation(fs, io, logger);
                FactoryResetLED = false
            } else {
                if (FactoryResetButtonPressedDuration % FACTORYRESETBLINKRATE) {
                    // If the LED is on, turn it off
                    //  of vice versa
                    if (FactoryResetLED) {
                        // Turn off LED
                        ledManager.OffOperation(fs, io, logger);
                        FactoryResetLED = false;
                    } else {
                        // Turn off LED
                        ledManager.WarningOperation(fs, io, logger);
                        FactoryResetLED = true;
                    }
                }
            }
            break;
    }
}

// This function performs the actual factory reset functions, 
//  the callback signals completion of the factory reset
function ProcessFactoryResetOperations(logger, callback) {

    logger.info('FactoryReset.ProcessFactoryResetOperations');

    // Copy over the factory default interfaces file to /etc/network
    exec("cp ./factory/interfaces /etc/network/interfaces", function (error, stdout, stderr) {
        if (error !== null) {
            logger.error('FactoryReset.ProcessFactoryResetOperations - cp ./factory/interfaces /etc/network/interfaces: error=' + error);
        }
        logger.info('FactoryReset.ProcessFactoryResetOperations - cp ./factory/interfaces /etc/network/interfaces: success=' + stdout);

        // Delete current user settings, which will then be replaced automatically
        //  by the factory default upon restart of Crest
        exec("cp ./factory/users.json ./users.json", function (error, stdout, stderr) {
            if (error !== null) {
                logger.error('FactoryReset.ProcessFactoryResetOperations - cp ./factory/users.json ./users.json: error=' + error);
            }
            logger.info('FactoryReset.ProcessFactoryResetOperations - cp ./factory/users.json ./users.json: success=' + stdout);

            // Delete current user settings, which will then be replaced automatically
            //  by the factory default upon restart of Crest
            exec("cp ./factory/ports.json ./data/ports.json", function (error, stdout, stderr) {
                if (error !== null) {
                    logger.error('FactoryReset.ProcessFactoryResetOperations - cp ./factory/ports.json ./data/ports.json: error=' + error);
                }
                logger.info('FactoryReset.ProcessFactoryResetOperations - cp ./factory/ports.json ./data/ports.json: success=' + stdout);

                // Delete the port events file(s)
                exec("rm ./log/port?events.json", function (error, stdout, stderr) {
                    if (error !== null) {
                        logger.error('FactoryReset.ProcessFactoryResetOperations - rm ./log/port?events.json: error=' + error);
                    }
                    logger.info('FactoryReset.ProcessFactoryResetOperations - rm ./log/port?events.json: success=' + stdout);

                    // Sync the results so the files are made persistent
                    exec("sync", function (error, stdout, stderr) {
                        if (error !== null) {
                            logger.error("FactoryReset.ProcessFactoryResetOperations: Error in sync after loading factory defaults. Error=" + error);
                        } else {
                            SummitUser = require("./users.json");
                            logger.info("FactoryReset.ProcessFactoryResetOperations: Successfully recovered factory defaults.");

                            // Call back to finish up our configuration changes
                            //  return back to calling API function
                            logger.info('FactoryReset.ProcessFactoryResetOperations: Calling FactoryResetReboot ... will execute in 500ms');
                            setTimeout(function () { FactoryResetReboot(logger) }, 500);
                        }
                    });
                });
            });
        });
    });
}

function FactoryResetReboot(logger) {

    // ***** very ungraceful exit *****
    logger.info('FactoryReset.FactoryResetReboot has been called!!!');
    exec("/sbin/shutdown -r now", function (error, stdout, stderr) {
        //logger.error('sync stderr: ' + stderr);
        if (error !== null) {
            logger.error('FactoryResetReboot - /sbin/shutdown -r now error=' + error);
        }
    });
}
